require 'open3'
require 'uri'

require 'liquid'

require 'specular/version'

class Specular
  # == Constants ============================================================

  STATUS_LABEL = {
    failed: 'Failed 🛑',
    passed: 'Passed ✅',
    skipped: 'Skipped ⚠️'
  }.freeze

  MESSAGE_EVENTS = %i[ default success failure ].freeze

  MESSAGES = {
    slack: {
      default: 'RSpec tests for {% if repo %}<{{repo}}|{{name}}>{% else %}{{name}}{% endif %} have {% if created_issues.github.url %}<{{created_issues.github.url}}|{{result}}>{% else %}{{result}}{% endif %}'
    },
    github: {
      default: 'RSpec test report ({{result}})'
    }
  }

  REF_DEFAULT = 'main'

  # == Extensions ===========================================================

  # == Exceptions ===========================================================

  class Error < StandardError; end

  # == Properties ===========================================================

  attr_reader :results

  # == Class Methods ========================================================

  # == Instance Methods =====================================================

  def initialize(results, **context)
    @results = results
    @context = context

    if (@context[:repo] and !@context[:repo].end_with?('/'))
      @context[:repo] = '%s/' % @context[:repo]
    end

    @context[:result] = self.failures? ? :failed : :passed

    @context[:messages] = MESSAGES.keys.to_h do |type|
      [
        type,
        MESSAGE_EVENTS.to_h do |event|
          [
            event,
            @context.dig(:messages, type, event) || MESSAGES.dig(type, event)
          ]
        end
      ]
    end
  end

  def git_ref
    @context[:ref] ||= Open3.popen3('git', 'rev-parse', 'HEAD') do |_sin, sout, _serr, _proc|
      sout.readlines.first.chomp
    end
  end

  def failures?
    @results.dig(:summary, :failure_count) > 0
  end

  def message(service: :github)
    Liquid::Template.parse(
      @context.dig(:messages, service, @context[:result]) ||
        @context.dig(:messages, service, :default)
    ).render(Specular::Support.stringify_keys(@context))
  end

  def examples
    if (@context.dig(:show, :all))
      @results[:examples]
    else
      @results[:examples].select do |example|
        example[:status] == 'failed'
      end
    end
  end

  def status_label(status)
    STATUS_LABEL[status.to_sym] || status
  end

  def source_link(file, line)
    if (@context[:repo])
      '[`%s:%d`](%s#L%d)' % [
        file,
        line,
        URI.join(
          @context[:repo] + 'blob/%s/' % [
            self.git_ref,
          ],
          file
        ),
        line
      ]
    else
      '`%s:%d`' % [
        file,
        line
      ]
    end
  end

  def result_format(result)
    return unless (result)

    "```\n%s%s```\n" % [
      result,
      result.end_with?("\n") ? "" : "\n"
    ]
  end

  def report(title: false)
    [
      *(title ? ("<!-- %s -->\n" % self.message(service: :github)) : ()),
      "## Test Results\n",
      self.examples.map do |example|
        case (example[:status])
        when 'failed'
          "### %s\n\n**Status:** %s\n**Source:** %s\n\n%s" % [
            example[:full_description],
            status_label(example[:status]),
            source_link(example[:file_path], example[:line_number]),
            result_format(example.dig(:exception, :message))
          ]
        else
          "### %s\n\n**Status:** %s\n**Source:** %s\n\n" % [
            example[:full_description],
            status_label(example[:status]),
            source_link(example[:file_path], example[:line_number])
          ]
        end
      end,
      "## Summary\n\n**Duration:** %.2fs\n**Failures:** %d\n" % [
        @results.dig(:summary, :duration),
        @results.dig(:summary, :failure_count)
      ],
      "## Git Context\n**Ref:** %s" % [
        self.git_ref
      ]
    ].join("\n")
  end

  def create_issue!(services = %i[ github ])
    @context[:created_issues] = services.to_h do |service|
      [
        service,
        Specular::Integration.for(service, @context).create_issue!(
          title: self.message(service: service),
          body: self.report
        )
      ]
    end
  end

  def post_message!(services = %i[ slack ])
    @context[:posted_messages] = services.to_h do |service|
      [
        service,
        Specular::Integration.for(service, @context).post_message!(
          message: self.message(service: service)
        )
      ]
    end
  end
end

require_relative 'specular/integration'
require_relative 'specular/support'
